/****************************************************************************
 *   Copyright (C) 2006-2013 by Jason Ansel, Kapil Arya, and Gene Cooperman *
 *   jansel@csail.mit.edu, kapil@ccs.neu.edu, gene@ccs.neu.edu              *
 *                                                                          *
 *  This file is part of DMTCP.                                             *
 *                                                                          *
 *  DMTCP is free software: you can redistribute it and/or                  *
 *  modify it under the terms of the GNU Lesser General Public License as   *
 *  published by the Free Software Foundation, either version 3 of the      *
 *  License, or (at your option) any later version.                         *
 *                                                                          *
 *  DMTCP is distributed in the hope that it will be useful,                *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU Lesser General Public License for more details.                     *
 *                                                                          *
 *  You should have received a copy of the GNU Lesser General Public        *
 *  License along with DMTCP:dmtcp/src.  If not, see                        *
 *  <http://www.gnu.org/licenses/>.                                         *
 ****************************************************************************/

#include <linux/sched.h>    /* Definition of struct clone_args */
#include <sched.h>          /* Definition of CLONE_* constants */
#include <sys/syscall.h>
#include <features.h>       /* for macro __GLIBC_PREREQ() */
#include "constants.h"
#include "dmtcpworker.h"
#include "mtcpinterface.h"
#include "syscallwrappers.h"
#include "dmtcp.h"
#include "uniquepid.h"
#include "util.h"
#include "../jalib/jassert.h"
#include "../jalib/jalloc.h"
#include "threadsync.h"
#include "processinfo.h"
#include "threadlist.h"
#include "siginfo.h"

#define PAGE_SHIFT      12
#define PAGE_SIZE       (1UL << PAGE_SHIFT)
#define PAGE_MASK       (~(PAGE_SIZE-1))

#define PAGE_ALIGN(addr)        (((addr)+PAGE_SIZE-1)&PAGE_MASK)

using namespace dmtcp;

static void *register_thread(void *arg);
static void *pthread_start(void *arg);
int clone_start(void *arg);

#ifdef CLONE3
/* The pthread_create wrapper receives start_routine and arg as parametrs.
 *   It then takes arg and creates a new thread_arg
 *   of type 'struct ThreadArg'.  The pthread_create wrapper then
 *   calls glibc:pthread_create(thread, attr, register_thread, thread_arg).
 *   Then thread_arg is passed down to pthread_start()
 *   which then unpacks it and calls start_routine on thread_arg.arg.
 *   The flow of control in the child thread is:
 * The pthread_create wrapper fnc calls this first:
 *     register_thread(thread_arg) -> clone_start(thread_arg) ->
 *        pthread_start(thread_arg) -> start_routine(arg)
 * FIXME:  We should simplify this into:
 *           register_thread(thread_arg) -> start_routine(arg) ->
 *         The current structure was based on the idea of interposing
 *           both on pthread_create and on clone, using some common code.
 *           But now that clone3 is a glibc private symbol and cannot be
 *           interposed on, this no longer makes sense.
 */
#endif
struct ThreadArg {
  union {
    int (*fn) (void *arg);
    void * (*pthread_fn) (void *arg); // pthread_create calls fn -> void *
  };
  void *arg;
  void *mtcpArg;
  pid_t virtualTid;
#ifdef CLONE3
  int (*clone_fn) (void *arg); // NOT USED
  pid_t ctid;
  void *start_routine_retval; // retval of start_routine of pthread_create
#endif
};

#ifdef CLONE3
LIB_PRIVATE
void *register_thread(void *thread_arg)
{
  // We were called by parent thread.  We need to make a copy of
  //   the thread_arg that we inherited, since the parent thread can
  //   finish its call frame.
  struct ThreadArg thread_arg_child = *(struct ThreadArg *)thread_arg;
  thread_arg_child.ctid = dmtcp_gettid();
  clone_start(&thread_arg_child);
  // Was allocated by JALLOC_HELPER_MALLOC in calling thread in pthread_create,
  //   before creating child thread.
  JALLOC_HELPER_FREE(thread_arg);
  return thread_arg_child.start_routine_retval;
}
#endif

// Invoked via __clone
LIB_PRIVATE
int clone_start(void *arg)
{
  Thread *thread = (Thread*) arg;

  ThreadSync::initThread();

  ThreadList::updateTid(thread);

  DmtcpWorker::eventHook(DMTCP_EVENT_THREAD_START, NULL);

  /* Thread finished initialization.  It's now safe for this thread to
   * participate in checkpoint.  Decrement the uninitializedThreadCount in
   * DmtcpWorker.
   */
  ThreadSync::decrementUninitializedThreadCount();

  JLOG(DMTCP)("Calling user function") (dmtcp_gettid());
#ifdef CLONE3
  pthread_start(arg);
  int ret = 0;  /* return 0 exit status for thread; FIXME: should return retval? */
#else
  int ret = thread->fn(thread->arg);
#endif

  ThreadList::threadExit();
  return ret;
}

//#ifndef CLONE3
/*****************************************************************************
 *
 *  This is our clone system call wrapper
 *
 *    Note:
 *
 *      pthread_create eventually calls __clone to create threads
 *      It uses flags = 0x3D0F00:
 *	      CLONE_VM = VM shared between processes
 *	      CLONE_FS = fs info shared between processes (root, cwd, umask)
 *	   CLONE_FILES = open files shared between processes (fd table)
 *	 CLONE_SIGHAND = signal handlers and blocked signals shared
 *	 			 (sigaction common to parent and child)
 *	  CLONE_THREAD = add to same thread group
 *	 CLONE_SYSVSEM = share system V SEM_UNDO semantics
 *	  CLONE_SETTLS = create a new TLS for the child from newtls parameter
 *	 CLONE_PARENT_SETTID = set the TID in the parent (before MM copy)
 *	CLONE_CHILD_CLEARTID = clear the TID in the child and do
 *				 futex wake at that address
 *	      CLONE_DETACHED = create clone detached
 *
 *****************************************************************************/
//need to forward user clone
extern "C" int __clone(int (*fn) (void *arg), void *child_stack, int flags,
                       void *arg, int *ptid,
                       struct user_desc *tls, int *ctid)
{
  WRAPPER_EXECUTION_DISABLE_CKPT();
  ThreadSync::incrementUninitializedThreadCount();

  Thread *thread = ThreadList::getNewThread();
  ThreadList::initThread(thread, fn, arg, flags, ptid, ctid);
//  if (ckpthread == NULL) {
//    ckptthread = thread;
//    thread->stateInit(ST_CKPNTHREAD);
//  }

  pid_t tid = _real_clone(clone_start, child_stack, flags, thread,
                          ptid, tls, ctid);

  if (tid == -1) {
    JLOG(DMTCP)("Clone call failed")(JASSERT_ERRNO);
    ThreadSync::decrementUninitializedThreadCount();
    ThreadList::threadIsDead(thread);
  } else {
    DmtcpWorker::eventHook(DMTCP_EVENT_THREAD_CREATED, NULL);
  }

  WRAPPER_EXECUTION_ENABLE_CKPT();
  return tid;
}
// #endif
#if 0
#if defined(__i386__) || defined(__x86_64__)
asm (".global clone ; .type clone,@function ; clone = __clone");
#elif defined(__arm__)
// In arm, '@' is a comment character;  Arm uses '%' in type directive
asm (".global clone ; .type clone,%function ; clone = __clone");
#else
# error Not implemented on this architecture
#endif
#endif


// NOT TESTED;  NOT USED UNTIL __clone3 becomes public.
//need to forward user clone3
//This was introduced in glibc-2.34, and pthread_create now calls clone3().
//  long syscall(SYS_clone3, struct clone_args *cl_args, size_t size);
//  'man clone3' omits last two args in Sept., 2021.
//  For all args, see:
//    glibc-2.34/sysdeps/unix/sysv/linux/clone-internal.c:__clone_internal()
//Unfortunately, in glibc-2.34, they document clone3 in the man page, and
//  they use it for themselves internally, but they don't play nice and make
//  it public for dynamic linking by other people, for interposing.  :-(
extern "C" long __clone3(struct clone_args *cl_args, size_t size,
                         int (*func) (void *arg), void *arg)
{
  uint64_t child_stack = (uint64_t)cl_args->stack + cl_args->stack_size;
  child_stack = PAGE_ALIGN(child_stack); 
  // dispatch to our own __clone wrapper, before going to glibc.
  // FIXME:  We should add our __clone wrapper logic here and call clone3.
  int rc = __clone(func, (void *)child_stack, cl_args->flags & ~0xff, arg,
                    (int *)cl_args->parent_tid,
                    (struct user_desc *)cl_args->tls,
                    (int *)cl_args->child_tid);
  return (long)rc;
}

void*
getDmtcpClone3Addr()
{
  return (void *)&__clone3;
}


// Invoked via pthread_create as start_routine
// On return, it calls mtcp_threadiszombie()
static void *pthread_start(void *arg)
{
  struct ThreadArg *threadArg = (struct ThreadArg*) arg;
  void *thread_arg = threadArg->arg;
  void * (*pthread_fn) (void *) = threadArg->pthread_fn;
  pid_t virtualTid = threadArg->virtualTid;

  JASSERT(pthread_fn != 0x0);
#ifndef CLONE3
  JALLOC_HELPER_FREE(arg); // Was allocated in calling thread in pthread_create
#endif

  // Unblock ckpt signal (unblocking a non-blocked signal has no effect).
  // Normally, DMTCP wouldn't allow the ckpt signal to be blocked. However, in
  // some situations (e.g., timer_create), libc would internally block all
  // signals before calling pthread_create to create a helper thread.  Since,
  // the child threads inherit parent signal mask, the helper thread has all
  // signals blocked.
  sigset_t set;
  sigaddset(&set, SigInfo::ckptSignal());
  JASSERT(_real_pthread_sigmask(SIG_UNBLOCK, &set, NULL) == 0) (JASSERT_ERRNO);

  ThreadSync::threadFinishedInitialization();
  void *result = (*pthread_fn)(thread_arg);
#ifdef CLONE3
  // This will be used by register_thread()
  ((struct ThreadArg*)thread_arg)->start_routine_retval = result;
#endif
  JLOG(DMTCP)("Thread returned") (virtualTid);
  WRAPPER_EXECUTION_DISABLE_CKPT();
  ThreadList::threadExit();
  /*
   * This thread has finished its execution, do some cleanup on our part.
   *  erasing the virtualTid entry from virtualpidtable
   *  FIXME: What if the process gets checkpointed after erase() but before the
   *  thread actually exits?
   */
  DmtcpWorker::eventHook(DMTCP_EVENT_PTHREAD_RETURN, NULL);
  WRAPPER_EXECUTION_ENABLE_CKPT();
  ThreadSync::unsetOkToGrabLock();
  return result;
}


extern "C" int pthread_create(pthread_t *thread, const pthread_attr_t *attr,
                              void *(*start_routine)(void*), void *arg)
{
#ifdef CLONE3
  // From our __clone() wrapper:
  // If glibc-2.34 or greater, we don't define our own __clone.
  // So, we do the bookkeeping of our own __clone() and clone_start() here.
  WRAPPER_EXECUTION_DISABLE_CKPT();
  ThreadSync::incrementUninitializedThreadCount();

  Thread *dmtcp_thread = ThreadList::getNewThread();
  typedef int (*clone_start_fn)(void *);
  int (*fn)(void*) = (clone_start_fn)start_routine;
  // FIXME:  Should fill in ctid.  Can we do this inside the new thread?
  /* Standard flags that glibc uses to create a new thread */
  int flags = CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND |
	      CLONE_THREAD | CLONE_SYSVSEM | CLONE_SETTLS |
	      CLONE_PARENT_SETTID | CLONE_CHILD_CLEARTID | CLONE_DETACHED;
  ThreadList::initThread(dmtcp_thread, fn, arg, flags,
		         NULL /* ptid */, NULL /* ctid */);
#endif
  int retval;
  // We have to use DMTCP-specific memory allocator because using glibc:malloc
  // can interfere with user threads.
  // We use JALLOC_HELPER_FREE to free this memory in two places:
  // 1. near the beginning of pthread_start (wrapper for start_routine),
  //     providing that the __clone call succeeds with no tid conflict.
  // 2. if the call to _real_pthread_create fails, then free memory
  //     near the end of this function.
  // We use MALLOC/FREE so that pthread_create() can be called again, without
  // waiting for the new thread to give up the buffer in pthread_start().
  struct ThreadArg *threadArg =
    (struct ThreadArg *) JALLOC_HELPER_MALLOC (sizeof (struct ThreadArg));
  threadArg->pthread_fn = start_routine;
  threadArg->arg = arg;

  /* pthread_create() should acquire the thread-creation lock. Not doing so can
   * result in a deadlock in the following scenario:
   * 1. user thread: pthread_create() - acquire wrapper-execution lock
   * 2. ckpt-thread: SUSPEND msg received, wait on wrlock for wrapper-execution lock
   * 3. user thread: __clone() - try to acquire wrapper-execution lock
   *
   * We also need to increment the uninitialized-thread-count so that it is
   * safe to checkpoint the newly created thread.
   *
   * There is another possible deadlock situation if we do not grab the
   * thread-creation lock:
   * 1. user thread: pthread_create(): waiting on tbl_lock inside libpthread
   * 2. ckpt-thread: SUSPEND msg received, wait on wrlock for wrapper-exec lock
   * 3. uset thread: a. exiting after returning from user fn.
   *                 b. grabs tbl_lock()
   *                 c. tries to call free() to deallocate previously allocated
   *                 space (stack etc.). The free() wrapper requires
   *                 wrapper-exec lock, which is not available.
   */
  bool threadCreationLockAcquired = ThreadSync::threadCreationLockLock();
  ThreadSync::incrementUninitializedThreadCount();
#ifdef CLONE3
  // This will do the DMTCP initialization of the thread,
  //   and then eventually call start_routine().
  // In glibc-2.34 and greater, we no longer wrap __clone.  So, we invoke
  //   our own clone_start() here.
  // The new child thread will invoke:
  //   register_thread() -> clone_start() -> pthread_start() -> start_routine()
  // intermixed with possible wrapper functions from glibc.
  retval = _real_pthread_create(thread, attr, register_thread, threadArg);
#else
  retval = _real_pthread_create(thread, attr, pthread_start, threadArg);
#endif
  if (threadCreationLockAcquired) {
    ThreadSync::threadCreationLockUnlock();
  }
  if (retval == 0) {
    ProcessInfo::instance().clearPthreadJoinState(*thread);
  } else { // if we failed to create new pthread
    JALLOC_HELPER_FREE(threadArg);
    ThreadSync::decrementUninitializedThreadCount();
  }
#ifdef CLONE3
  // From our __clone() wrapper:
  // FIXME:  We could get the tid by passing a pointer arg to pthread_start()
  //   and then have our wrapper function write its tid at a known location.
  //   For now, we cheat and find tid via glibc:nptl/descr.h: 'struct pthread'.
  // WARNING:  The number 24, below, may change w/ GLIBC version.
  char *addr = (char *)thread + 24*sizeof(void *) +
	       2*sizeof(void *); /* 2*sizeof(void *) == sizeof(list_t) */
  pid_t tid = *(pid_t *)addr;
  if (tid == -1) {
    JLOG(DMTCP)("Clone call failed")(JASSERT_ERRNO);
    ThreadSync::decrementUninitializedThreadCount();
    ThreadList::threadIsDead(dmtcp_thread);
  } else {
    DmtcpWorker::eventHook(DMTCP_EVENT_THREAD_CREATED, NULL);
  }

  WRAPPER_EXECUTION_ENABLE_CKPT();
#endif
  return retval;
}

extern "C" void pthread_exit(void * retval)
{
  WRAPPER_EXECUTION_DISABLE_CKPT();
  ThreadList::threadExit();
  DmtcpWorker::eventHook(DMTCP_EVENT_PTHREAD_EXIT, NULL);
  WRAPPER_EXECUTION_ENABLE_CKPT();
  ThreadSync::unsetOkToGrabLock();
  _real_pthread_exit(retval);
  for (;;); // To hide compiler warning about "noreturn" function
}

/*
 * pthread_join() is a blocking call that waits for the given thread to exit.
 * It examines the value of 'tid' field in 'struct pthread' of the given
 * thread. The kernel will write '0' to this field when the thread exits.
 *
 * In pthread_join(), the thread makes a futex call in the following fashion:
 *   _tid = pd->tid;
 *   while !succeeded
 *     futex(&pd->tid, FUTEX_WAIT, 0, _tid, ...)
 * As we can see, if the checkpoint is issued during pthread_join(), on
 * restart, the tid would have changed, but the call to futex would still used
 * the previously cached tid. This causes the caller to spin with 100% cpu
 * usage.
 *
 * The fix is to use the non blocking pthread_tryjoin_np function. To maintain
 * the semantics of pthread_join(), we need to ensure that only one thread is
 * allowed to wait on the given thread. This is done by keeping track of
 * threads that are being waited on by some other thread.
 *
 * Similar measures are taken for pthread_timedjoin_np().
 */
static struct timespec ts_100ms = {0, 100 * 1000 * 1000};
extern "C" int pthread_join(pthread_t thread, void **retval)
{
  int ret;
  struct timespec ts;
  if (!ProcessInfo::instance().beginPthreadJoin(thread)) {
    return EINVAL;
  }

  while (1) {
    WRAPPER_EXECUTION_DISABLE_CKPT();
    ThreadSync::unsetOkToGrabLock();
    JASSERT(clock_gettime(CLOCK_REALTIME, &ts) != -1);
    TIMESPEC_ADD(&ts, &ts_100ms, &ts);
    ret = _real_pthread_timedjoin_np(thread, retval, &ts);
    WRAPPER_EXECUTION_ENABLE_CKPT();
    ThreadSync::setOkToGrabLock();
    if (ret != ETIMEDOUT) {
      break;
    }
  }

  ProcessInfo::instance().endPthreadJoin(thread);
  return ret;
}

extern "C" int pthread_tryjoin_np(pthread_t thread, void **retval)
{
  int ret;
  if (!ProcessInfo::instance().beginPthreadJoin(thread)) {
    return EINVAL;
  }

  WRAPPER_EXECUTION_DISABLE_CKPT();
  ret = _real_pthread_tryjoin_np(thread, retval);
  WRAPPER_EXECUTION_ENABLE_CKPT();

  ProcessInfo::instance().endPthreadJoin(thread);
  return ret;
}

extern "C" int pthread_timedjoin_np(pthread_t thread, void **retval,
                                    const struct timespec *abstime)
{
  int ret;
  struct timespec ts;
  if (!ProcessInfo::instance().beginPthreadJoin(thread)) {
    return EINVAL;
  }

  /*
   * We continue to call pthread_tryjoin_np (and sleep) until we have gone past
   * the abstime provided by the caller
   */
  while (1) {
    WRAPPER_EXECUTION_DISABLE_CKPT();
    JASSERT(clock_gettime(CLOCK_REALTIME, &ts) != -1);
    if (TIMESPEC_CMP(&ts, abstime, <)) {
      TIMESPEC_ADD(&ts, &ts_100ms, &ts);
      ret = _real_pthread_timedjoin_np(thread, retval, &ts);
    } else {
      ret = ETIMEDOUT;
    }
    WRAPPER_EXECUTION_ENABLE_CKPT();

    if (ret == EBUSY || ret == 0) {
      break;
    }
    if (TIMESPEC_CMP(&ts, abstime, >=)) {
      ret = ETIMEDOUT;
      break;
    }
  }

  ProcessInfo::instance().endPthreadJoin(thread);
  return ret;
}
